#include <stdio.h> // for file I/O
#include <stdlib.h> // for exit()
#include <string.h> // for strncmp()
#include <memory.h> // for memset()
#include <math.h> // for pow()

#define LCD_WIDTH 112
#define LCD_HEIGHT 64

// defaults

#define PREROLL 10 // leadoff frames to let the phases drift apart
#define FRAMES 34 // frames for a still

#define DECAY 0.001 // exponential decay: new = old + decay*diff

// todo: determine by histogram?
#define GAMMA 0.5 // gamma correction
#define LUMA_MIN 16  // offset
#define LUMA_MAX 235 // maximum seen

#define FPS_PLAY 67.0 // the playback speed we reach
#define FPS_MOVIE 29.97 // original framerate of the movie


// command line configuration: what shall we do?
struct
{
	char* szInFile;
    char* szOutFile;
    double dMovieFPS;
    double dPlayFPS;
    double dGamma;
    int nLumaMin;
    int nLumaMax;
    int nFrames;
} gCfg;



typedef struct
{
    double pix[LCD_WIDTH][LCD_HEIGHT];
} tFrame;


// fill array with a constant
void fill_array(tFrame* pArray, double dValue)
{
    int x,y;
    for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            pArray->pix[x][y] = dValue;
        }
    }
}


// add noise to a frame
void noise_array(tFrame* pArray, double level)
{
    int x,y;
    for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            pArray->pix[x][y] += level * (double)rand() / (double)RAND_MAX;
            pArray->pix[x][y] -= level / 2;
        }
    }
}


// load a still picture
void read_still(tFrame* pIdeal, FILE* pFile, unsigned* pHistogram)
{
    unsigned char byte;
    int x,y;

    for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            fread(&byte, 1, 1 ,pFile);
            pHistogram[byte]++;
            pIdeal->pix[x][y] = (double)byte / 255.0; // range 0...1
        }
    }
}


// load a new frame of the movie, returns 0 if OK
int read_frame(FILE* pFile, tFrame* pNew, double* pLookup, unsigned* pHistogram)
{
    unsigned char byte;
    int x,y;
	unsigned char frame[LCD_WIDTH*LCD_HEIGHT];
	unsigned char* pRun = frame;

    if (fread(frame, sizeof(frame), 1, pFile) != 1)
		return 1; // eof

	for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            byte = *pRun++;
            pHistogram[byte]++;
            pNew->pix[x][y] = pLookup[byte]; // gamma corrected, etc.
        }
    }
    fseek(pFile, LCD_WIDTH*LCD_HEIGHT/2, SEEK_CUR); // skip UV
    return 0;
}


// build a lookup table with the float value for each byte
void InitLookup(double* pLookup)
{
	int i;
	double value;

	for (i=0; i<256; i++)
	{
        value = (double)i;
        value -= gCfg.nLumaMin; // subtract offset
        value /= (gCfg.nLumaMax - gCfg.nLumaMin); // range 0...1
        if (value < 0.0) value = 0.0;
        if (value > 1.0) value = 1.0;
        value = pow(value, gCfg.dGamma); // gamma corrected
		pLookup[i] = value;
	}
}


// load a new frame into the halftoning
int change_frame(tFrame* pNew, tFrame* pIdeal, tFrame* pCurrent)
{
    int x,y;
    double value;
    double diff;

    for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            value = pNew->pix[x][y];
            diff = value - pIdeal->pix[x][y];
            pIdeal->pix[x][y] = value;
            pCurrent->pix[x][y] += (1-DECAY)*diff; // not full diff to make change more likely
        }
    }
    return 0;
}


// the heart of the halftoning
void process(tFrame* pSet, tFrame* pCurrent, tFrame* pIdeal, double decay)
{
    int x,y;

    for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            pSet->pix[x][y] = (pIdeal->pix[x][y] < pCurrent->pix[x][y]) ? 0.0 : 1.0; // driven value
            pCurrent->pix[x][y] += (pSet->pix[x][y] - pCurrent->pix[x][y]) * decay; // IIR filter
        }
    }
}


// write a frame in Rockbox format
void write_binary(FILE* pFile, tFrame* pSet)
{
    int x, y;
    unsigned bit;
    unsigned char byte;
	unsigned char frame[LCD_WIDTH*(LCD_HEIGHT/8)];
	unsigned char* pRun = frame;

    for (y=0; y<LCD_HEIGHT; y+=8)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            byte = 0;
            for (bit=0; bit<8; bit++)
            {
                if (pSet->pix[x][y+bit] < 0.5)
                {   // dark pixel
                    byte |= (0x01 << bit);
                }
            }
			*pRun++ = byte;
        }
    }
    fwrite(frame, sizeof(frame), 1, pFile);
}


// write a frame in RAW format
void write_gray(FILE* pFile, tFrame* pSet)
{
    int x, y;
    unsigned char byte;

    for (y=0; y<LCD_HEIGHT; y++)
    {
        for (x=0; x<LCD_WIDTH; x++)
        {
            byte = (unsigned char)(pSet->pix[x][y] * 255.0);
            fwrite(&byte, 1, 1, pFile);
        }
    }
}


// main loop for video encoding
int encode_video(FILE* pInFile, FILE* pOutFile, double* pLookup, unsigned* pHistogram)
{
    double frametime = 0;
    
    tFrame ideal; 
    tFrame next; 
    tFrame current; 
    tFrame set; // driven value (0 or 1)

    fill_array(&ideal, 0.0);
    fill_array(&current, 0.0);
    noise_array(&current, 0.5/255);

    while(1)
    {
        if (frametime >= 0.0)
        {
            if (read_frame(pInFile, &next, pLookup, pHistogram))
                break;
            frametime -= 1.0 / gCfg.dMovieFPS;
            change_frame(&next, &ideal, &current);
        }
        process(&set, &current, &ideal, DECAY);
        write_binary(pOutFile, &set);
        //write_gray(pOutFile, &set);
        frametime += 1.0 / gCfg.dPlayFPS;
    }

    return 0;
}


// main loop for still frame encoding
int encode_still(FILE* pInFile, FILE* pOutFile, unsigned* pHistogram)
{
    tFrame ideal; 
    tFrame current; 
    tFrame set; // driven value (0 or 1)
    int i;

    read_still(&ideal, pInFile, pHistogram);
    current = ideal;
    noise_array(&current, 0.5/255);

    // let it settle
    for (i=0; i<PREROLL; i++)    
    {
        process(&set, &current, &ideal, DECAY);
    }

    // now do the real thing
    for (i=0; i<gCfg.nFrames; i++)    
    {
        process(&set, &current, &ideal, DECAY);
        write_binary(pOutFile, &set);
    }
    return 0;
}


// diagnostic function
void print_histogram(unsigned* pHistogram)
{
    unsigned sum = 0;
    int i;

    for (i=0; i<256; i++)
    {
        sum += pHistogram[i];
    }

    printf("Input file histogram:\n");
    for (i=0; i<256; i++)
    {
        printf("p[%03d] = %.8lf\n", i, (double)pHistogram[i]/(double)sum);
    }

}

//#define PROFILING

int ProcessCmdLine(int argc, char* argv[])
{
	char* szValue;
    argc--; // exclude our name
	argv++;
#ifdef PROFILING
	gCfg.szInFile = "c:\\UserData\\hohensoh\\AJB\\rockbox\\experimental\\halftone\\doom3.yuv";
	gCfg.szOutFile = "c:\\UserData\\hohensoh\\AJB\\rockbox\\experimental\\halftone\\doom3.rvf";
	return 0;
#else
	if (argc < 2)
	{
		printf("Usage: halftone <infile> <outfile> [-option {parameter}]\n");
		printf("  Creates a Rockbox video file from still or movie.\n");
		printf("  Still image input format: RAW grayscale 112*64 (7168 bytes)\n");
		printf("  Video input format: YUV 4:2:0 planar (from avitoyuv.exe), 112*64\n");
		printf("Possible options are: (one letter for options is enough)\n");
		printf("-movie_fps <fps>   FPS of the input file, default is %.2f\n", FPS_MOVIE);
		printf("-play_fps <fps>   FPS of Rockbox playback, default is %.2f\n", FPS_PLAY);
		printf("-gamma <gamma>   Gamma value to use, default 0.5 for video, 1.0 for stills \n");
		printf("-offset <min>   Minimum luma value of video file, default = %d\n", LUMA_MIN);
		printf("-saturation <max>   Maximum luma value to use, default = %d\n", LUMA_MAX);
		printf("-frames <count>   Frames for a still picture, default = %d\n", FRAMES);
		printf("\n");
		printf("Examples: \n");
		printf("halftone doom.yuv doom.rvf\n");
		printf("  convert video with default settings\n");
		printf("halftone doom.yuv doom.rvf -g 0.5 -o 25 -s 218\n");
		printf("  convert video with the used values\n");
		printf("halftone lena.raw lena.rvf -g 1 -o 0 -s 255\n");
		printf("  convert still image with neutral settings, full range\n");

		exit (0);
	}
#endif

    gCfg.szInFile = *argv++;
    argc--;

    gCfg.szOutFile = *argv++;
    argc--;

	while (argc)
	{
		if (!strncmp("-movie_fps", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.dMovieFPS = atof(szValue);
		}
		else if (!strncmp("-play_fps", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.dPlayFPS = atof(szValue);
		}
		else if (!strncmp("-gamma", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.dGamma = atof(szValue);
		}
		else if (!strncmp("-offset", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.nLumaMin = atoi(szValue);
		}
		else if (!strncmp("-saturation", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.nLumaMax = atoi(szValue);
		}
		else if (!strncmp("-frames", *argv, 2))
		{
			szValue = *++argv;
			if (--argc <= 0 || **argv == '-')
			{	
				printf("No argument given for option %s, aborting.\n", argv[-1]);
				exit (-2);
			}
            gCfg.nFrames = atoi(szValue);
		}
		else
		{
			printf("Unknown option %s, aborting. Use 'halftone' without options for help.\n", *argv);
			exit(-1);
		}

		argv++;
		argc--;
	}
	
	return 0;
}


int main (int argc, char** argv)
{
    FILE* pInFile;
    FILE* pOutFile;
    long filesize;
    unsigned histogram[256];
    int frameclocks;

    memset(histogram, 0, sizeof(histogram));
    memset(&gCfg, 0, sizeof(gCfg));
    // set defaults
    gCfg.dMovieFPS = FPS_MOVIE;
    gCfg.dPlayFPS = FPS_PLAY;
    gCfg.dGamma = GAMMA;
    gCfg.nLumaMin = LUMA_MIN;
    gCfg.nLumaMax = LUMA_MAX;
    gCfg.nFrames = FRAMES;

    ProcessCmdLine(argc, argv); // what to do

    // round to Archos timer resolution
    frameclocks = (int)(11059200.0/gCfg.dPlayFPS) / 4;
    gCfg.dPlayFPS = 11059200.0 / (double)frameclocks / 4.0;

    pInFile = fopen(gCfg.szInFile, "rb");
    if (pInFile == NULL)
    {
        printf("cannot open input file %s\n", gCfg.szInFile);
        exit(2);
    }

    pOutFile = fopen(gCfg.szOutFile, "wb");
    if (pOutFile == NULL)
    {
        printf("cannot open output file %s, gCfg.szOutFile)\n");
        exit(3);
    }

    fseek(pInFile, 0, SEEK_END);
    filesize = ftell(pInFile);
    fseek(pInFile, 0, SEEK_SET);

    if (filesize <= LCD_WIDTH*LCD_HEIGHT)
        encode_still(pInFile, pOutFile, histogram);
    else
	{
		double lookup[256];
		InitLookup(lookup); // make a table for gamma/luma
        encode_video(pInFile, pOutFile, lookup, histogram);
	}

    fclose(pInFile);
    fclose(pOutFile);

    print_histogram(histogram);

    return 0;
}
